#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <libndls.h>

#define MEMORY_START_SIZE 50
#define DEBUG 1


extern "C"
{
	#ifdef DEBUG
		static uint32_t malloc_count = 0; // should be 0 after all is freed
	#endif
	static void** memory_allocations = NULL;
	static uint32_t memory_allocations_size = 0;
	void* __real_malloc(size_t size);
	void __real_free(void* p);

	void* __wrap_malloc(size_t size)
	{
		void* block = __real_malloc(size);
		if (block == NULL)
		{
			return NULL;
		}
		#ifdef DEBUG
			printf("malloc called, block address: %p\n",block);
			fflush(stdout);
		#endif
		if (memory_allocations == NULL)
		{
			#ifdef DEBUG
				puts("allocating pointer array\n");
				fflush(stdout);
			#endif
			memory_allocations = (void**) __real_malloc(sizeof(void*)*MEMORY_START_SIZE);
			if (memory_allocations == NULL)
			{
				__real_free(block);
				return NULL;
			}
			#ifdef DEBUG
				malloc_count++;
			#endif
			memory_allocations_size = MEMORY_START_SIZE;
			for (uint32_t i = 0;i<memory_allocations_size;i++)
			{
				memory_allocations[i] = NULL;
			}
		}
		for (uint32_t i = 0;i<memory_allocations_size;i++)
		{
			if (memory_allocations[i] == NULL)
			{
				#ifdef DEBUG
					malloc_count++;
				#endif
				memory_allocations[i] = block;
				return block;
			}
		}
		#ifdef DEBUG
			puts("resizing pointer array\n");
			fflush(stdout);
		#endif
		// no free entry found, make the array bigger
		void** new_memory_allocations = (void**) __real_malloc(sizeof(void*)*memory_allocations_size*2);
		if (new_memory_allocations == NULL)
		{
			__real_free(block);
			return NULL;
		}
		for (uint32_t i = 0;i<memory_allocations_size*2;i++)
		{
			new_memory_allocations[i] = NULL;
		}
		for (uint32_t i = 0;i<memory_allocations_size;i++)
		{
			new_memory_allocations[i] = memory_allocations[i];
		}
		__real_free(memory_allocations);
		memory_allocations = new_memory_allocations;
		memory_allocations_size *= 2;
		for (uint32_t i = 0;i<memory_allocations_size;i++)
		{
			if (memory_allocations[i] == NULL)
			{
				#ifdef DEBUG
					malloc_count++;
				#endif
				memory_allocations[i] = block;
				return block;
			}
		}
		puts("no place found after resizing!\n");
		fflush(stdout);
		// should never come to this
		__real_free(block);
		return NULL;
	}
	void __wrap_free(void* p)
	{
		#ifdef DEBUG
			puts("free called!");
			fflush(stdout);
		#endif
		for (uint32_t i = 0;i<memory_allocations_size;i++)
		{
			if (memory_allocations[i] == p)
			{
				#ifdef DEBUG
					printf("block freed: %p\n",p);
					fflush(stdout);
				#endif
				memory_allocations[i] = NULL;
				break;
			}
		}
		#ifdef DEBUG
			malloc_count--;
		#endif
		__real_free(p);
	}
	// this function is used in the inline assembly statement below, but GCC doesn't recognize that and gives a warning
	#pragma GCC diagnostic ignored "-Wunused-function"
	// WARNING this breaks all assumptions C++ has, so it's probably best to call it in an inline assembly statement and call __crt0_exit in the next instruction
	// deallocate_exit does that
	static void deallocate_all()
	{
		for (uint32_t i = 0;i<memory_allocations_size;i++)
		{
			if (memory_allocations[i] != NULL)
			{
				#ifdef DEBUG
					printf("freeing before exit: %p\n",memory_allocations[i]);
					malloc_count--;
				#endif
				__real_free(memory_allocations[i]);
			}
		}
		if (memory_allocations != NULL)
		{
			__real_free(memory_allocations);
			memory_allocations = NULL;
			#ifdef DEBUG
				malloc_count--;
			#endif
		}
		memory_allocations_size = 0;
		#ifdef DEBUG
			printf("malloc count: %lu\n",malloc_count);
			fflush(stdout);
		#endif
	}
	#pragma GCC diagnostic pop
	void __attribute__ ((noreturn)) deallocate_exit(int code)
	{
		register int32_t c asm("r0") = code;
		asm volatile(
		"bl deallocate_all \n"
		"bl __crt0_exit"::"r" (c));
		__builtin_unreachable();
	}
}